import * as Twitter from 'twitter';
import TwitterTypes from 'twitter-d';

import { getLogger } from './loggers';
import QQBot from './koishi';
import { chainPromises, BigNumOps } from './utils';

interface IWorkerOption {
  bot: QQBot;
  privateAuthToken: string;
  privateCsrfToken: string;
}

interface ISearchQueryConfig {
  query: string;
  count?: number;
  since?: string;
  until?: string;
}

export let doSearch = (
  conf: {[key in keyof ISearchQueryConfig]: string},
  receiver: IChat
): void => {
  throw Error();
};

const TWITTER_EPOCH = 1288834974657;
const snowflake = (epoch: number) => Number.isNaN(epoch) ? undefined :
  BigNumOps.lShift(String(epoch - 1 - TWITTER_EPOCH), 22);

const logger = getLogger('twitter');

export type FullUser = TwitterTypes.FullUser;
export type Entities = TwitterTypes.Entities;
export type ExtendedEntities = TwitterTypes.ExtendedEntities;
export type MediaEntity = TwitterTypes.MediaEntity;

interface Tweet extends TwitterTypes.Status {
  user: FullUser;
  retweeted_status?: Tweet;
}

namespace AdaptiveSearch {
  export interface Response {
    globalObjects: {
      tweets: {[key: number]: Tweet & {user_id: number, user_id_str: string}};
      users: {[key: number]: FullUser};
      topics: {[key: number]: {[key: string]: any}};
      [key: string]: any;
    };
    timeline: {
      id: string;
      instructions: [{
        addEntries: {entries: TimelineEntry[]};
      }];
    };
  }
  
  export interface TimelineEntry {
    entryId: string;
    sortIndex: string;
    content: {item: object} | {operation: object};
  }
  
  export interface TimelineCursor extends TimelineEntry {
    content: {
      operation: {cursor: {value: string}}
    }
  };
  
  export const isCursor = (entry: TimelineEntry): entry is TimelineCursor =>
    entry.sortIndex === '0' || entry.sortIndex === '999999999';
}

export default class {

  private privateClient: Twitter;
  private bot: QQBot;

  constructor(opt: IWorkerOption) {
    this.privateClient = new Twitter({
      bearer_token: 'AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA',
      request: {
        headers: {
          'Content-Type': 'application/x-www-form-urlencoded',
          Cookie: `auth_token=${opt.privateAuthToken}; ct0=${opt.privateCsrfToken};`,
          'X-CSRF-Token': opt.privateCsrfToken,
        },
      }
    } as any);
    this.bot = opt.bot;
    doSearch = ({query, count, since, until}, receiver) => {
      const countNum = Number(count) || 15;
      this.search({
        query,
        count: Math.abs(countNum),
        since: BigNumOps.parse(since) || snowflake(new Date(since).getTime()),
        until: BigNumOps.parse(until) || snowflake(new Date(until).getTime()),
      })
        .then(tweets => chainPromises(
          tweets.map(tweet => () => this.bot.sendTo(receiver, `\
用户：${tweet.user.name} (@${tweet.user.screen_name})
编号：${tweet.id_str}
时间：${tweet.created_at}
媒体：${tweet.extended_entities ? '有' : '无'}
正文：\n${tweet.full_text.replace(/^([\s\S\n]{50})[\s\S\n]+?( https:\/\/t.co\/.*)?$/, '$1…$2')}`
          ))
            .concat(() => this.bot.sendTo(receiver, tweets.length ?
              '自定义查询完毕，使用 /twitter_view <编号> 查看推文详细内容。' :
              '自定义查询完毕，没有找到符合条件的推文。'
            ))
        ))
        .catch((err: {code: number, message: string}[]) => {
          logger.warn(`error retrieving timeline: ${err[0]?.message || err}`);
          return this.bot.sendTo(receiver, `自定义查询时出现错误：${err[0]?.message || err}`);
        });
    };
  }

  public search = (conf: ISearchQueryConfig) => {
    for (const key in conf) {
      if (!conf[key]) delete conf[key];
    }
    const origQuery = conf.query;
    delete conf.query;
    const paramNames = {until: 'max_id', since: 'since_id'};
    logger.info(`performing custom query ${origQuery} with config: ${JSON.stringify(conf)}`);
    const query = origQuery.replace(
      new RegExp(`(${Object.values(paramNames).join('|')}):\d+`, 'g'),
      ($0, $1) => conf[$1] ? '' : $0
    ).replace(/\(\)/g, '') +
      Object.keys(paramNames).map(
        k => conf[k] && `(${paramNames[k]}:${conf[k]})`
      ).join();
    const doSearch = (q: string, tweets: Tweet[] = [], cursor?: string): Promise<Tweet[]> =>
      this.privateClient.get('https://api.twitter.com/2/search/adaptive.json', {
        q,
        tweet_mode: 'extended',
        count: 20,
        include_entities: true,
        query_source: 'typed_query',
        ...(cursor && {cursor}),
      }).then(({
        globalObjects: {tweets: tweetDict, users: userDict},
        timeline: {instructions: {0: {addEntries: {entries}}}}
      }: AdaptiveSearch.Response) => {
        const newTweets = Object.values(tweetDict);
        let bottomCursor: AdaptiveSearch.TimelineCursor;
        if (newTweets.length) {
          logger.debug(`fetched tweets: ${JSON.stringify(newTweets)}`);
          entries.sort(
            (e1, e2) => Number(e1.sortIndex) - Number(e2.sortIndex)
          )
          if (AdaptiveSearch.isCursor(entries[0])) {
            bottomCursor = entries[0];
            logger.info(`custom query ${origQuery} yielded ${
              newTweets.length
            } new tweets, next query will follow tweet ${
              entries[1].entryId.replace(/^sq-I-t-/, '')
            }`);
          } else {
            logger.info(`custom query ${origQuery} ended after yielding ${
              newTweets.length
            } new tweets, last entry was tweet ${
              entries[0].entryId.replace(/^sq-I-t-/, '')
            }`);
          }
          newTweets.forEach(tweet => Object.assign(tweet, {
            user: userDict[tweet.user_id_str],
          }));
          tweets.push(...newTweets);
        }
        if (!bottomCursor || !newTweets.length || tweets.length >= conf.count) {
          logger.info(`custom query ${origQuery} finished successfully, ${
            tweets.length
          } tweets have been fetched`);
          return tweets.slice(0, conf.count);
        }
        return doSearch(query, tweets, bottomCursor?.content.operation.cursor.value);
      });
    return doSearch(query);
  };
}
